@@ -6,7 +6,7 @@
  * An encapsulation layer for msghdr structures, cause that's the worst
  * system-level API I have ever seen.
  *
- * $Id: msghdr.c,v 1.15 2002/10/15 23:20:48 dougk Exp $
+ * $Id: msghdr.c,v 1.16 2003/02/04 18:50:04 dougk Exp $
  */
  
 #include "../config.h"
@@ -232,7 +232,7 @@ void msg_addChar(message_t *msg, char c) {
 }
 
 void msg_addString(message_t *msg, const char *s) {
-    size_t len = strlen(s);
+    size_t len = (s != NULL ? strlen(s) : 0);
     msg_addInt(msg, len);
     msg_addData(msg, s, len);
 }
@@ -410,6 +410,8 @@ int msg_recvmsg(message_t *msg, int fd)
 {
     /* Nuke the contents of the current msg.  Make sure we have enough
      * space for most sane things...
+     *
+     * Also, be robust against EINTR.
      */
     struct msghdr sys_msg;
     int rval;
@@ -430,7 +432,9 @@ int msg_recvmsg(message_t *msg, int fd)
 	cmsglen, 0
     };
 
-    rval = recvmsg(fd, &sys_msg, 0);
+    do {
+        rval = recvmsg(fd, &sys_msg, 0);
+    } while (rval == -1 && errno == EINTR);
 
     msg->recv_len = rval;
     return rval;
