@@ -6,7 +6,7 @@
  * Provides the server half of the process.  This half retains priviledge
  * and should be nei-invulnerable.
  *
- * $Id: privman.cc,v 1.54 2002/11/12 04:43:51 dougk Exp $
+ * $Id: privman.cc,v 1.56 2003/03/08 06:02:58 dougk Exp $
  */
 #include "../config.h"
 #include <stdio.h>
@@ -39,6 +39,10 @@
 #include <sys/resource.h>
 #endif
 
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
+
 #include <set>
 #include <map>
 
@@ -66,6 +70,57 @@ static bool     p_wait_on_child = true;
  */
 config_t               *config;
 
+
+/* Have to wrap wait.  Go ahead, ask why... :) */
+/* "object" data for mywait and the signal function. */
+static int mywait_status_v[4];
+static int mywait_pid_v[4];
+static struct rusage mywait_rusage_v[4];
+static int mywait_i = 0;
+#define sizeof_array(a) (sizeof(a) / sizeof(a[0]))
+
+static void sigchld_handler(int UNUSED(i)) {
+    do { 
+        mywait_pid_v[mywait_i] = wait4(-1, mywait_status_v + mywait_i,
+                                            WNOHANG,
+                                            mywait_rusage_v + mywait_i);
+        if (mywait_pid_v[mywait_i] <= 0 )
+            break;
+        mywait_i = (mywait_i + 1) % sizeof_array(mywait_status_v);
+    } while (1);
+}
+/* Overhead for sigaction */
+static struct sigaction child_sigaction;
+
+/* No support for options. */
+static pid_t mywait4(pid_t pid, int *status, int options, struct rusage *usage)
+{
+    unsigned i;
+    int rc;
+    if (pid < -1) {
+        errno = EINVAL;
+        return -1;
+    }
+    do {
+        for (i = 0 ; i < sizeof_array(mywait_status_v) ; ++i) {
+            if (mywait_pid_v[i] < 1)
+                continue; /* skip this one. */
+            if (pid == -1 || mywait_pid_v[i] == pid) {
+                if (status != NULL)
+                    *status = mywait_status_v[i];
+                if (usage != NULL)
+                    *usage = mywait_rusage_v[i];
+                rc = mywait_pid_v[i];
+
+                mywait_pid_v[i] = 0;
+                return rc;
+            }
+        }
+    } while (!(options & WNOHANG) &&
+            select(0,NULL,NULL,NULL,NULL) == -1 && errno == EINTR);
+    return -1;
+}
+
 /* Dispatch table */
 static std::map< enum commands, void(*)(message_t*) > function_map;
 
@@ -433,6 +488,13 @@ static void pamStart(message_t *msg)
     service = msg_getAllocStr(msg, 128, "pamStart: bad service");
     user    = msg_getAllocStr(msg, 128, "pamStart: bad user");
 
+    /* msghdr combines "" and NULL.  pam_start doens't like "", 
+     * so make sure it gets NULL.
+     */
+    if (user[0] == '\0') {
+        free(user);
+        user = NULL;
+    }
     pconv = (struct pam_conv){ convert_punt, 0};
 
     retval = pam_start(service, user, &pconv, &handle);
@@ -442,7 +504,9 @@ static void pamStart(message_t *msg)
     msg_addPtr(msg, handle);
 
     msg_sendmsg(msg, privmand_fd, "pamStart(sendmsg)");
-    free(service); free(user);
+    free(service);
+    if (user != NULL)
+        free(user);
 }
 
 static void pamAuthenticate(message_t *msg)
@@ -572,7 +636,7 @@ static void privWait4(message_t *msg) {
     if (flags & WANTS_RUSAGE)
         r = &ruse;
 
-    pid = wait4(pid, s, options, r);
+    pid = mywait4(pid, s, options, r);
 
     msg_clear(msg);
 
@@ -919,6 +983,12 @@ PAM_SIMPLE_HANDLER(pamFailDelay,     (int(*)(pam_handle_t*,int))pam_fail_delay)
 
 void privman_serv_init(void)
 {
+    struct sigaction old;
+    /* Set up signal handler. */
+    child_sigaction.sa_handler = sigchld_handler;
+    child_sigaction.sa_flags = SA_NOCLDSTOP;
+    sigaction(SIGCHLD, &child_sigaction, &old);
+
     function_map[ CMD_OPEN              ] = openFile;
     function_map[ CMD_UNLINK            ] = unlinkFile;
     function_map[ CMD_BIND              ] = bindPort;
@@ -964,7 +1034,7 @@ void privman_serv_init(void)
         if (p_wait_on_child) {
             /* Other side closed pipe */
             int status;
-            waitpid(child_pid, &status, 0);
+            mywait4(child_pid, &status, 0, 0);
             if (WIFEXITED(status))
                 _exit(WEXITSTATUS(status));
             else {
