@@ -6,7 +6,7 @@
  * This program executes "id" as user "joe".  Its simply a small test
  * program for priv_execve.
  *
- * $Id: myexec.c,v 1.5 2002/10/30 20:49:39 dougk Exp $
+ * $Id: myexec.c,v 1.6 2002/11/01 05:39:58 dougk Exp $
  */
 
 #include "../config.h"
@@ -15,7 +15,9 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/types.h>
 #include <unistd.h>
+#include <pwd.h>
 
 #include "privman.h"
 
@@ -24,11 +26,20 @@ extern char **environ;
 int main(void)
 {
     char *argv[2] = { "id", 0 };
+    struct passwd *pw;
+
     priv_init("myexec");
 
-    priv_execve(ID_PATH, argv, environ, "joe", 0);
+    priv_execve(ID_PATH, argv, environ, "root", 0);
+    perror("priv_execve (should fail)");
 
-    perror("priv_execve");
+    pw = getpwent();
+    while (pw != NULL && pw->pw_uid < 10)
+        pw = getpwent();
+
+    priv_execve(ID_PATH, argv, environ, pw->pw_name, 0);
+
+    perror("priv_execve(should not have failed)");
 
     return -1;
 }
